import sys
from itertools import chain
from pathlib import Path

# This package is only required when updating the emissions GWP conversion factors
import globalwarmingpotentials as gwp  # type: ignore

# Base path for package code
BASE_PATH = Path(__file__).parent

# Base path for package data
DATA_PATH = BASE_PATH / "data"


# Format strings for emissions()
_EMI_HEADER = """# This file was generated using:
#    python -m iam_units.update emissions
# DO NOT ALTER THIS FILE MANUALLY!
"""

# Format string for individual metrics. To expand the set of supported
# conversions, duplicate and modify the first pair of lines in the context.
# Only the portion before of the ':' needs to be modified. Currently supported:
# 1. Mass
# 2. Mass rate, i.e. mass per time.
# 3. Flux, i.e. mass per area per time.
_EMI_DATA = f"""{_EMI_HEADER}
@context(_a=NaN) {{metric}}
    [mass] -> [_GWP]: value * (_a * _gwp / kg)
    [_GWP] -> [mass]: value / (_a * _gwp / kg)
    [mass] / [time] -> [_GWP] / [time]: value * (_a * _gwp / kg)
    [_GWP] / [time] -> [mass] / [time]: value / (_a * _gwp / kg)
    [mass] / [time] / [area] -> [_GWP] / [time] / [area]: value * (_a * _gwp / kg)
    [_GWP] / [time] / [area] -> [mass] / [time] / [area]: value / (_a * _gwp / kg)


    {{defs}}
@end
"""  # noqa: E501

# Format string for an importable Python module defining the *pattern* regex,
# which resembles: (?<=[ -])(CO2|C|N2O|CH4)(?=[ -/]|[^\w]|$)
# - Preceded by a space or '-' character.
# - Followed by a space, '-', '/', end-of-string, or non-word (\w) character.
#   The latter avoids matching only the 'C' within 'CH4'.
_EMI_CODE = rf"""{_EMI_HEADER}
import re

GWP_VERSION = '{gwp.__version__}'

# All available metrics usable with convert_gwp().
METRICS = [
    '{{metrics}}'
]

# All recognised emission species usable with convert_gwp(). See *pattern*.
SPECIES = [
    '{{symbols}}',
    ]

# Sets of symbols that refer to the same species and are interchangeable.
EQUIV = [
    set({{equiv}}),
    ]

# Regular expression for one *SPECIES* in a pint-compatible unit string.
pattern = re.compile(
    '(?<=[ -])('
    + '|'.join(SPECIES)
    + r')(?=[ -/]|[^\w]|$)')
"""

# Format string for list of metrics.
_EMI_METRICS = f"""{_EMI_HEADER}

# Define contexts for each set of metrics

{{metrics}}
"""

# Equivalents: different symbols for the same species.
_EMI_EQUIV = {
    "CO2": {
        "CO2_eq": None,
        "CO2e": None,
        "CO2eq": None,
        "C": "44. / 12 * ",
        "Ce": "44. / 12 * ",
    }
}


def emissions():
    """Update emissions definitions files."""
    data_path = DATA_PATH / "emissions"

    # Import data from `globalwarmingpotentials`, get list of species aka symbols.
    data = gwp.as_frame().sort_index()
    symbols = data.index

    # Format and write the species defs file
    lines = [_EMI_HEADER]
    for (species, alias) in _EMI_EQUIV.items():
        lines.extend(
            f"a_{a} = {factor or ''}a_{species}" for a, factor in alias.items()
        )
    lines.extend(f"a_{s} = NaN" for s in symbols)
    lines.append("")
    (data_path / "species.txt").write_text("\n".join(lines))

    # Write a Python module with a regex matching the species names

    # Prepare list including all symbols
    all_alias_groups = list([key, *value] for key, value in _EMI_EQUIV.items())
    all_symbols = list(chain(*all_alias_groups, symbols))

    # Format and write `emissions.py`
    code = _EMI_CODE.format(
        metrics="',\n    '".join(list(data.columns)),
        symbols="',\n    '".join(all_symbols),
        equiv="),\n    set(".join(map(repr, all_alias_groups)),
    )
    (BASE_PATH / "emissions.py").write_text(code)

    # Format and write `metrics.txt"`
    code = _EMI_METRICS.format(
        metrics="\n".join([f"@import {m}.txt" for m in data.columns])
    )
    (data_path / "metrics.txt").write_text(code)

    # Write one file containing a context for each metric
    for metric in data.columns:
        # Conversion factor definitions
        defs = [
            f"a_{species} = {value}" for species, value in data[metric].dropna().items()
        ]

        # Format the template with the definitions
        content = _EMI_DATA.format(metric=metric, defs="\n    ".join(defs))

        # Write to file
        (data_path / f"{metric}.txt").write_text(content)


if __name__ == "__main__":
    # Invoked using 'python -m iam_units.update'
    # For each additional argument, call the function of the same name
    for module in sys.argv[1:]:
        locals()[module]()



# This file was generated using:
#    python -m iam_units.update emissions
# DO NOT ALTER THIS FILE MANUALLY!

import re

GWP_VERSION = "0.9.3"

# All available metrics usable with convert_gwp().
METRICS = [
    "SARGWP100",
    "AR4GWP100",
    "AR5GWP100",
    "AR5CCFGWP100",
    "AR6GWP100",
    "AR6GWP20",
    "AR6GWP500",
    "AR6GTP100",
]

# All recognised emission species usable with convert_gwp(). See *pattern*.
SPECIES = [
    "CO2",
    "CO2_eq",
    "CO2e",
    "CO2eq",
    "C",
    "Ce",
    "C10F18",
    "C2F6",
    "C3F8",
    "C4F10",
    "C5F12",
    "C6F14",
    "C7F16",
    "C8F18",
    "CCl4",
    "CF4",
    "CFC11",
    "CFC113",
    "CFC114",
    "CFC115",
    "CFC12",
    "CFC13",
    "CH2Cl2",
    "CH3Br",
    "CH3CCl3",
    "CH3Cl",
    "CH4",
    "CHCl3",
    "HCFC123",
    "HCFC124",
    "HCFC141b",
    "HCFC142b",
    "HCFC21",
    "HCFC22",
    "HCFC225ca",
    "HCFC225cb",
    "HCFE235da2",
    "HFC125",
    "HFC134",
    "HFC134a",
    "HFC143",
    "HFC143a",
    "HFC152",
    "HFC152a",
    "HFC161",
    "HFC227ea",
    "HFC23",
    "HFC236cb",
    "HFC236ea",
    "HFC236fa",
    "HFC245ca",
    "HFC245fa",
    "HFC32",
    "HFC365mfc",
    "HFC41",
    "HFC4310mee",
    "HFE125",
    "HFE134",
    "HFE143a",
    "HFE227ea",
    "HFE236ca12",
    "HFE236ea2",
    "HFE236fa",
    "HFE245cb2",
    "HFE245fa1",
    "HFE245fa2",
    "HFE263fb2",
    "HFE329mcc2",
    "HFE338mcf2",
    "HFE338pcc13",
    "HFE347mcc3",
    "HFE347mcf2",
    "HFE347pcf2",
    "HFE356mec3",
    "HFE356pcc3",
    "HFE356pcf2",
    "HFE356pcf3",
    "HFE365mcf3",
    "HFE374pc2",
    "HFE4310pccc124",
    "HFE569sf2",
    "Halon1201",
    "Halon1202",
    "Halon1211",
    "Halon1301",
    "Halon2402",
    "N2O",
    "NF3",
    "PFPMIE",
    "SF5CF3",
    "SF6",
    "SO2F2",
    "cC3F6",
    "cC4F8",
]

# Sets of symbols that refer to the same species and are interchangeable.
EQUIV = [
    set(["CO2", "CO2_eq", "CO2e", "CO2eq", "C", "Ce"]),
]

# Regular expression for one *SPECIES* in a pint-compatible unit string.
pattern = re.compile("(?<=[ -])(" + "|".join(SPECIES) + r")(?=[ -/]|[^\w]|$)")



from pathlib import Path

import pint
from pint.formatting import format_unit
from pint.util import to_units_container

from . import emissions

__all__ = [
    "convert_gwp",
    "format_mass",
    "registry",
]


# Package registry using definitions.txt
registry = pint.UnitRegistry()
registry.load_definitions(str(Path(__file__).parent / "data" / "definitions.txt"))


def convert_gwp(metric, quantity, *species):
    """Convert *quantity* between GHG *species* with a GWP *metric*.

    Parameters
    ----------
    metric : str or None
        Metric conversion factors to use. May be :obj:`None` if the input and
        output species are the same.
        Use :code:`iam_units.emissions.METRICS` for a list of available metrics.
    quantity : str or pint.Quantity or tuple
        Quantity to convert. If a tuple of (magnitude, unit), these are passed
        as arguments to :class:`pint.Quantity`.
    species : sequence of str, length 1 or 2
        Output, or (input, output) species symbols, e.g. ('CH4', 'CO2') to
        convert mass of CH₄ to GWP-equivalent mass of CO₂. If only the output
        species is provided, *quantity* must contain the symbol of the input
        species in some location, e.g. '1.0 tonne CH4 / year'.

    Returns
    -------
    pint.Quantity
        `quantity` converted from the input to output species.

    Notes
    -----
    The conversion factors are taken from the `globalwarmingpotentials` package,
    see https://github.com/openclimatedata/globalwarmingpotentials.
    You can use :code:`iam_units.emissions.GWP_VERSION` to check from which version
    of that package the conversion tables in the `iam_units` package were generated.
    """
    # Handle *species*: either (in, out) or only out
    try:
        species_in, species_out = species
    except ValueError:
        if len(species) != 1:
            raise ValueError("Must provide (from, to) or (to,) species")
        species_in, species_out = None, species[0]

    # Split *quantity* if it is a tuple. After this step:
    # - *mag* is the magnitude, or None.
    # - *expr* is a string expression for either just the units, or the entire
    #   quantity, including magnitude, as a str or pint.Quantity.
    mag, expr = quantity if isinstance(quantity, tuple) else (None, quantity)

    # If species_in wasn't provided, then *expr* must contain it
    if not species_in:
        # Extract it using the regex, then re-assemble the expression for the
        # units or whole quantity
        q0, species_in, q1 = emissions.pattern.split(expr, maxsplit=1)
        expr = q0 + q1

    # *metric* can only be None if the input and output species symbols are
    # identical or equivalent
    if metric is None:
        if species_in == species_out or any(
            {species_in, species_out} <= g for g in emissions.EQUIV
        ):
            metric = "AR5GWP100"
        elif species_in in species_out:
            # Eg. 'CO2' in 'CO2 / a'. This is both a DimensionalityError and a
            # ValueError (no metric); raise the former for pyam compat
            raise pint.DimensionalityError(species_in, species_out)
        else:
            msg = f"Must provide GWP metric for ({species_in}, {species_out})"
            raise ValueError(msg)

    # Ensure a pint.Quantity object:
    # - If *quantity* was a tuple, use the 2-arg constructor.
    # - If a str, use the 1-arg form to parse it.
    # - If already a pint.Quantity, this is a no-op.
    args = (expr,) if mag is None else (mag, expr)
    quantity = registry.Quantity(*args)

    # Construct intermediate units with the same dimensionality as *quantity*,
    # except '[mass]' replaced with the dummy unit '_gwp'
    dummy = quantity.units / registry.Unit("tonne / _gwp")

    # Convert to dummy units using 'a' for the input species; then back to the
    # input units using 'a' for the output species.
    return quantity.to(dummy, metric, _a=f"a_{species_in}").to(
        quantity.units, metric, _a=f"a_{species_out}"
    )


def format_mass(obj, info, spec=None):
    """Format the units of *obj* with *info* inserted after its mass unit.

    Parameters
    ----------
    obj : pint.Quantity or pint.Unit
    info : str
        Any information, e.g. the symbol of a GHG species.
    spec : str, optional
        Pint formatting specifier such as "H" (HTML format), "~C" (compact format with
        symbols), etc.
    """
    spec = spec or obj.default_format

    try:
        # Use only the units of a Quantity object
        obj = obj.units
    except AttributeError:
        pass  # Already a Unit object

    # Use the symbol if the modifier "~" is in `spec`; else the canonical name. cf.
    # pint.Unit.__format__()
    method = registry._get_symbol if "~" in spec else lambda k: k
    # Collect the pieces of the unit expression
    units = [[method(key), value] for key, value in obj._units.items()]

    # Index of the mass component
    mass_index = list(obj.dimensionality.keys()).index("[mass]")
    # Append the information (e.g. species) to the mass component
    units[mass_index][0] += f" {info}"

    # Prepare pint.util.UnitsContainer and hand off to pint's formatting. Discard "~"
    # (used above) and ":" (invalid in pint ≥0.18).
    return format_unit(
        to_units_container(dict(units), registry=registry),
        spec.replace("~", "").lstrip(":"),
    )
